package protoexample;

enum FOO {X=17;};

message Test {
   required string label = 1;
   optional int32 type = 2[default=77];
   repeated int64 reps = 3;
   optional group OptionalGroup = 4{
     required string RequiredField = 5;
   }
}
